"""WebSocket Gabriel client used to communicate with a Gabriel server."""

import asyncio
import logging
from typing import Callable
from urllib.parse import urlparse

import websockets
import websockets.client
from gabriel_protocol import gabriel_pb2

from gabriel_client.gabriel_client import (
    GabrielClient,
    InputProducer,
    TokenPool,
)

URI_FORMAT = "ws://{host}:{port}"


logger = logging.getLogger(__name__)
websockets_logger = logging.getLogger(websockets.__name__)

# The entire payload will be printed if this is allowed to be DEBUG
websockets_logger.setLevel(logging.INFO)


class WebsocketClient(GabrielClient):
    """A Gabriel client that talks to the server over WebSockets.

    The Gabriel server must be configured to use Websockets for client
    communication.
    """

    def __init__(
        self,
        server_endpoint: str,
        input_producers: list[InputProducer],
        consumer: Callable[[gabriel_pb2.Result], None],
        prometheus_port: int = 8001,
    ):
        """Initialize the client.

        Args:
        server_endpoint (str):
            The server endpoint to connect to. Must have the form
            'ws://interface:port'.
        input_producers (List[InputProducer]):
            A list of instances of InputProducer for the inputs
            produced by this client
        consumer (Callable[[gabriel_pb2.Result], None]):
            Callback for results from server
        prometheus_port (int):
            Port for Prometheus metrics.

        """
        super().__init__(prometheus_port)
        self.consumer = consumer
        self.input_producers = set(input_producers)

        self.server_endpoint = server_endpoint
        parsed_server_endpoint = urlparse(server_endpoint.lower())
        if parsed_server_endpoint.scheme != "ws":
            raise ValueError(
                f"Unsupported protocol {parsed_server_endpoint.scheme}"
            )

    async def launch_async(self):
        """Launch the client asynchronously."""
        async with websockets.connect(
            self.server_endpoint,
            max_size=None,
        ) as websocket:
            self._websocket = websocket
            consumer_task = asyncio.create_task(self._consumer_handler())
            tasks = [
                asyncio.create_task(self._producer_handler(input_producer))
                for input_producer in self.input_producers
            ]
            tasks.append(consumer_task)

            try:
                done, pending = await asyncio.wait(
                    tasks, return_when=asyncio.FIRST_COMPLETED
                )
            except asyncio.CancelledError:
                for task in tasks:
                    task.cancel()
                await asyncio.gather(*tasks, return_exceptions=True)
                raise
            for task in pending:
                task.cancel()
            logger.info("Disconnected From Server")

    async def _consumer_handler(self):
        while self._running:
            try:
                raw_input = await self._websocket.recv()
            except websockets.exceptions.ConnectionClosed:
                return  # stop the handler
            logger.debug("Received input from server")

            to_client = gabriel_pb2.ToClient()
            to_client.ParseFromString(raw_input)

            if to_client.HasField("welcome"):
                self._process_welcome(to_client.welcome)
            elif to_client.HasField("result_wrapper"):
                self._process_response(to_client.result_wrapper)
            else:
                raise Exception("Empty to_client message")

    def _process_welcome(self, welcome):
        self._num_tokens_per_producer = welcome.num_tokens_per_producer
        self._engine_ids = welcome.engine_ids
        self._welcome_event.set()

    def _process_response(self, result_wrapper):
        result = result_wrapper.result
        if result.status.code == gabriel_pb2.StatusCode.SUCCESS:
            self.record_response_latency(result_wrapper)
            self.consumer(result_wrapper.result)
        elif result.status.code == gabriel_pb2.StatusCode.NO_ENGINE_FOR_INPUT:
            raise Exception("No engine for input")
        else:
            status = gabriel_pb2.StatusCode.Name(result.status.code)
            logger.error("Output status was: %s", status)

        if result_wrapper.return_token:
            self._tokens[result_wrapper.producer_id].return_token()

    async def _producer_handler(self, producer: InputProducer):
        """Generate inputs and sends them to the server.

        Loop waiting until there is a token available. Then call
        producer to get the gabriel_pb2.InputFrame to send.

        Args:
            producer (InputProducer):
                The InputProducer instance that produces inputs for
                this client

        """
        await self._welcome_event.wait()
        token_pool = TokenPool(
            self._num_tokens_per_producer, producer.producer_id
        )
        self._tokens[producer.producer_id] = token_pool
        frame_id = 1

        while self._running:
            await token_pool.get_token()

            input_frame = await producer.produce()
            if input_frame is None:
                token_pool.return_token()
                logger.info("Received None from producer")
                continue

            from_client = gabriel_pb2.FromClient()
            from_client.frame_id = frame_id
            frame_id += 1
            from_client.producer_id = producer.producer_id

            target_engines = set(producer.get_target_engines())
            available_engines = set(self._engine_ids)

            if not target_engines.issubset(available_engines):
                msg = (
                    f"Attempt to target engines that are not connected "
                    f"to the server: {target_engines - available_engines}"
                )
                logger.error(msg)
                raise ValueError(msg)

            from_client.target_engine_ids.extend(producer.get_target_engines())
            from_client.input_frame.CopyFrom(input_frame)

            # Send input to server
            logger.debug(
                f"Sending input to server; producer={producer.producer_id}"
            )
            try:
                await self._send_from_client(from_client)
            except websockets.exceptions.ConnectionClosed:
                return  # stop the handler

    async def _send_from_client(self, from_client):
        self.record_send_metrics(from_client)
        # Removing this method will break measurement_client
        await self._websocket.send(from_client.SerializeToString())



"""ZeroMQ Gabriel client used to communicate with a Gabriel server."""

import asyncio
import logging
import time
from collections.abc import Iterable
from typing import Callable
from urllib.parse import urlparse

import zmq
import zmq.asyncio
from gabriel_protocol import gabriel_pb2
from google.protobuf.message import DecodeError

from gabriel_client.gabriel_client import (
    GabrielClient,
    InputProducer,
    TokenPool,
)

logger = logging.getLogger(__name__)


# The duration of time in seconds after which the server is considered to be
# disconnected.
SERVER_TIMEOUT_SECS = 10

HEARTBEAT = b""
# The interval in seconds at which a heartbeat is sent to the server
HEARTBEAT_INTERVAL = 1

# Message sent to the server to register this client
HELLO_MSG = b"Hello message"


class ZeroMQClient(GabrielClient):
    """A Gabriel client that talks to the server over ZeroMQ.

    The Gabriel server must be configured to use ZeroMQ for client
    communication.
    """

    def __init__(
        self,
        server_endpoint: str,
        input_producers: Iterable[InputProducer],
        consumer: Callable[[gabriel_pb2.Result], None],
        prometheus_port: int = 8001,
    ):
        """Initialize the client.

        Args:
        server_endpoint (str):
            The server endpoint to connect to. Must have the form
            'protocol://interface:port'. Protocols supported are
            tcp and ipc
        input_producers (Iterable[InputProducer]):
            An iterable of instances of InputProducer for the inputs
            produced by this client
        consumer (Callable[[gabriel_pb2.Result], None]):
            Callback for results from server
        prometheus_port (int):
            Port for Prometheus metrics.

        """
        super().__init__(prometheus_port)
        # Socket used for communicating with the server
        self._ctx = zmq.asyncio.Context()
        self._sock = self._ctx.socket(zmq.DEALER)
        self._sock.setsockopt(zmq.LINGER, 0)
        # Check whether IPC mode is enabled, and only use the host if so

        parsed_server_endpoint = urlparse(server_endpoint.lower())
        if parsed_server_endpoint.scheme not in ("tcp", "ipc"):
            raise ValueError(
                f"Unsupported protocol {parsed_server_endpoint.scheme}"
            )
        self._server_endpoint = server_endpoint

        self.input_producers = set(input_producers)
        self.consumer = consumer
        # Whether the client is connected to the server
        self._connected = asyncio.Event()
        # Indicates that a heartbeat was sent to the server but a heartbeat
        # wasn't received back from the server
        self._pending_heartbeat = False
        # The last time a heartbeat was sent to the server
        self._last_heartbeat_time = 0
        # Set by tasks to schedule a heartbeat to be sent to the server
        self._schedule_heartbeat = asyncio.Event()

        self._connected.set()

    def remove_input_producer(self, input_producer):
        """Remove an input producer from the client."""
        if input_producer not in self.input_producers:
            return False
        self.input_producers.remove(input_producer)
        return True

    async def launch_async(self):
        """Launch async tasks for running the client.

        Handles producing inputs, consuming results, and sending heartbeats to
        the server. Also sends a hello message to the server to register this
        client with the server.
        """
        logger.info(f"Connecting to server at {self._server_endpoint}")
        self._sock.connect(self._server_endpoint)

        await self._sock.send(HELLO_MSG)
        logger.info("Sent hello message to server")

        tasks = [
            asyncio.create_task(self._producer_handler(input_producer))
            for input_producer in self.input_producers
        ]
        tasks.append(asyncio.create_task(self._consumer_handler()))
        tasks.append(asyncio.create_task(self._heartbeat_loop()))

        try:
            await asyncio.gather(*tasks)
        except asyncio.CancelledError:
            for task in tasks:
                task.cancel()
            await asyncio.gather(*tasks, return_exceptions=True)
            raise
        except Exception as e:
            logger.error(f"Client encountered exception: {e}")
            for task in tasks:
                task.cancel()
            await asyncio.gather(*tasks, return_exceptions=True)
            raise
        finally:
            self._sock.close()
            self._ctx.term()

    async def _consumer_handler(self):
        """Handle messages from the server."""
        while self._running:
            # Wait for a message with a timeout, specified in milliseconds
            poll_result = await self._sock.poll(SERVER_TIMEOUT_SECS * 1000)
            if (poll_result & zmq.POLLIN) == 0:
                if self._connected.is_set():
                    logger.info("Disconnected from server")
                    self._connected.clear()
                else:
                    logger.info(
                        "Still disconnected; reconnecting and resending "
                        "heartbeat"
                    )

                # Resend heartbeat in case it was lost
                self._sock.close()
                self._sock = self._ctx.socket(zmq.DEALER)
                self._sock.setsockopt(zmq.LINGER, 0)
                self._sock.connect(self._server_endpoint)

                # Send heartbeat even though we are disconnected
                await self._send_heartbeat(True)
                continue

            raw_input = await self._sock.recv()

            if not self._connected.is_set():
                logger.info("Reconnected to server")
                self._connected.set()
                # Reset tokens for all producers
                for token_pool in self._tokens.values():
                    token_pool.reset_tokens()

            if raw_input == HEARTBEAT:
                logger.debug("Received heartbeat from server")
                self._pending_heartbeat = False
                continue

            logger.debug("Received message from server")

            to_client = gabriel_pb2.ToClient()
            try:
                to_client.ParseFromString(raw_input)
            except DecodeError as e:
                logger.error(f"Failed to decode message from server: {e}")
                continue

            if to_client.HasField("welcome"):
                logger.info("Received welcome from server")
                self._process_welcome(to_client.welcome)
            elif to_client.HasField("result_wrapper"):
                logger.debug("Processing response from server")
                self._process_response(to_client.result_wrapper)
            elif to_client.HasField("control"):
                logger.debug("Received control message from server")
                self._engine_ids = to_client.control.engine_ids
                logger.debug(f"Updating engine ids to: {self._engine_ids}")
            else:
                logger.critical(
                    "Fatal error: empty to_client message received from server"
                )
                raise Exception("Empty to_client message")

    def _process_welcome(self, welcome):
        """Process a welcome message received from the server.

        Args:
            welcome:
                The gabriel_pb2.ToClient.Welcome message received from
                the server

        """
        self._num_tokens_per_producer = welcome.num_tokens_per_producer
        self._engine_ids = welcome.engine_ids
        self._welcome_event.set()
        logger.info(
            f"Available engines: {self._engine_ids}; "
            f"number of tokens per producer: {self._num_tokens_per_producer}"
        )

    def _process_response(self, result_wrapper):
        """Process a result received from the server.

        Args:
            result_wrapper:
                The gabriel_pb2.ToClient.ResultWrapper message received from
                the server
        """
        result = result_wrapper.result
        result_status = result.status
        code = result_status.code
        msg = result_status.message
        if code == gabriel_pb2.StatusCode.SUCCESS:
            self.record_response_latency(result_wrapper)
            try:
                self.consumer(result)
            except Exception as e:
                logger.error(f"Error processing response from server: {e}")
                raise
        elif code == gabriel_pb2.StatusCode.NO_ENGINE_FOR_INPUT:
            logger.critical(f"Fatal error: no engine for input: {msg}")
            raise Exception(f"No engine for input: {msg}")
        elif code == gabriel_pb2.StatusCode.SERVER_DROPPED_FRAME:
            logger.error(
                f"Server {result.target_engine_id} dropped frame from "
                f"producer {result_wrapper.producer_id}: {msg}"
            )
        else:
            status_name = gabriel_pb2.StatusCode.Name(code)
            logger.error(
                f"Input from producer {result_wrapper.producer_id} targeting "
                f"server {result.target_engine_id} caused error "
                f"{status_name}: {msg}"
            )

        if result_wrapper.return_token:
            producer_id = result_wrapper.producer_id
            self._tokens[producer_id].return_token()
            logger.debug(
                f"Returning token for producer {producer_id}, total tokens "
                f"{self._tokens[producer_id].get_remaining_tokens()}"
            )

    async def _producer_handler(self, producer: InputProducer):
        """Generate inputs and sends them to the server.

        Loop waiting until there is a token available. Then call
        producer to get the gabriel_pb2.InputFrame to send.

        Args:
            producer (InputProducer):
                The InputProducer instance that produces inputs for
                this client

        """
        await self._welcome_event.wait()

        # Async task used to producer an input
        producer_task = None
        frame_id = 1

        producer_id = producer.producer_id
        token_pool = TokenPool(self._num_tokens_per_producer, producer_id)
        self._tokens[producer_id] = token_pool

        try:
            while self._running and producer in self.input_producers:
                if not producer.is_running():
                    logger.info(
                        f"Producer {producer.producer_id} is not running; "
                        f"waiting"
                    )
                    await producer.wait_for_running()
                    logger.info(f"Producer {producer.producer_id} resumed")

                try:
                    # Wait for a token. Time out to send a heartbeat to the
                    # server.
                    await asyncio.wait_for(
                        token_pool.get_token(), timeout=HEARTBEAT_INTERVAL
                    )
                except (TimeoutError, asyncio.TimeoutError):
                    self._schedule_heartbeat.set()
                    continue

                target_engines = producer.get_target_engines()
                if not target_engines:
                    logger.error(
                        f"{producer.producer_name} targets no engines"
                    )

                    token_pool.return_token()
                    raise ValueError(
                        f"{producer.producer_name} targets no engines"
                    )
                    continue

                # Stop producing inputs if disconnected from the server
                if not self._connected.is_set():
                    logger.debug("Stopping producer task")
                    if producer_task is not None:
                        producer_task.cancel()
                        producer_task = None
                    await self._connected.wait()
                    logger.debug("Resuming producer task")

                if producer_task is None:
                    logger.debug("Created new producer task")
                    producer_task = asyncio.create_task(producer.produce())
                try:
                    # Wait for the producer to produce an input. Time out to
                    # send a heartbeat to the server. Use an asyncio.shield to
                    # prevent task cancellation.
                    input_frame = await asyncio.wait_for(
                        asyncio.shield(producer_task),
                        timeout=HEARTBEAT_INTERVAL,
                    )
                except (TimeoutError, asyncio.TimeoutError):
                    token_pool.return_token()
                    self._schedule_heartbeat.set()
                    await asyncio.sleep(0)
                    continue

                logger.debug("Got input from producer")

                producer_task = None

                if input_frame is None:
                    token_pool.return_token()
                    logger.debug("Received None from producer")
                    continue
                if not input_frame.SerializeToString():
                    token_pool.return_token()
                    logger.error("Input producer produced empty frame")
                    continue

                from_client = gabriel_pb2.FromClient()
                from_client.frame_id = frame_id
                frame_id += 1
                from_client.producer_id = producer.producer_id

                target_engines = set(producer.get_target_engines())
                available_engines = set(self._engine_ids)

                if not target_engines.issubset(available_engines):
                    msg = (
                        f"Attempt to target engines that are not connected "
                        f"to the server: {target_engines - available_engines}"
                    )
                    logger.error(msg)
                    raise ValueError(msg)

                from_client.target_engine_ids.extend(
                    producer.get_target_engines()
                )
                from_client.input_frame.CopyFrom(input_frame)

                # Send input to server
                logger.debug(
                    f"Sending input to server; producer={producer.producer_id}"
                )
                await self.send_to_server(from_client)

                logger.debug(
                    "Semaphore for %s is %s",
                    producer.producer_id,
                    "LOCKED" if token_pool.is_locked() else "AVAILABLE",
                )
        finally:
            if producer_task is not None:
                producer_task.cancel()
                asyncio.gather(producer_task, return_exceptions=True)

    async def send_to_server(self, from_client: gabriel_pb2.FromClient):
        """Send a frame to the server."""
        self.record_send_metrics(from_client)
        await self._sock.send(from_client.SerializeToString())

    async def _send_heartbeat(self, force=False):
        """Send a heartbeat to the server.

        Args:
            force:
                Send a heartbeat even if disconnected from the server or
                if sufficient time has not passed since the last
                heartbeat was sent

        """
        # Do not send a heartbeat if disconnected from server or a heartbeat
        # is pending, unless force is True
        if not force and (
            not self._connected.is_set() or self._pending_heartbeat
        ):
            return
        logger.debug("Sending heartbeat to server")
        self._pending_heartbeat = True
        self._last_heartbeat_time = time.monotonic()
        await self._sock.send(HEARTBEAT)

    async def _heartbeat_loop(self):
        """Run the client heartbeat loop.

        Send a heartbeat when one is scheduled by another task and
        sufficient time has passed since the last heartbeat was sent.
        """
        while self._running:
            if not self._schedule_heartbeat.is_set():
                # Wait for heartbeat to be scheduled by another task
                await self._schedule_heartbeat.wait()
            self._schedule_heartbeat.clear()

            # Heartbeat was sent recently
            if (
                time.monotonic() - self._last_heartbeat_time
                < HEARTBEAT_INTERVAL
            ):
                continue
            await self._send_heartbeat()



"""Adapter to integrate OpenCV with Gabriel client framework."""

import logging

import cv2
import numpy as np
from gabriel_protocol import gabriel_pb2

from gabriel_client.gabriel_client import InputProducer

logger = logging.getLogger(__name__)


class OpencvAdapter:
    """Adapter to integrate OpenCV with the Gabriel client framework."""

    def __init__(
        self,
        preprocess,
        produce_extras,
        consume_frame,
        video_capture,
        engine_name,
    ):
        """Initialize the adapter.

        Args:
            preprocess: A function to preprocess the video frame.
            produce_extras:
                A function to produce extra fields for the input frame.
            consume_frame:
                A function to consume the output frame from the server.
            video_capture: The OpenCV video capture object.
            engine_name: The name of the video processing engine.

        """
        self._preprocess = preprocess
        self._produce_extras = produce_extras
        self._consume_frame = consume_frame
        self._video_capture = video_capture
        self._engine_name = engine_name

    def get_producer_wrappers(self):
        """Get the producer wrappers for the video source."""

        async def producer():
            _, frame = self._video_capture.read()
            if frame is None:
                return None

            frame = self._preprocess(frame)
            _, jpeg_frame = cv2.imencode(".jpg", frame)

            input_frame = gabriel_pb2.InputFrame()
            input_frame.payload_type = gabriel_pb2.PayloadType.IMAGE
            input_frame.payloads.append(jpeg_frame.tobytes())

            extras = self._produce_extras()
            if extras is not None:
                input_frame.extras.Pack(extras)

            return input_frame

        return [
            InputProducer(
                producer=producer, target_engine_ids=[self._engine_name]
            )
        ]

    def consumer(self, result_wrapper):
        """Consume the output frame from the server."""
        if len(result_wrapper.results) != 1:
            logger.error(
                "Got %d results from server", len(result_wrapper.results)
            )
            return

        result = result_wrapper.results[0]
        if result.payload_type != gabriel_pb2.PayloadType.IMAGE:
            type_name = gabriel_pb2.PayloadType.Name(result.payload_type)
            logger.error("Got result of type %s", type_name)
            return

        np_data = np.frombuffer(result.payload, dtype=np.uint8)
        frame = cv2.imdecode(np_data, cv2.IMREAD_COLOR)

        self._consume_frame(frame, result_wrapper.extras)



"""Abstract base class for a Gabriel client and related classes."""

import asyncio
import logging
import threading
import time
import uuid
from abc import ABC, abstractmethod
from collections.abc import Callable, Coroutine, Iterable
from typing import Any, Union

from gabriel_protocol.gabriel_pb2 import FromClient, InputFrame, ToClient
from prometheus_client import Counter, Gauge, Histogram

logger = logging.getLogger(__name__)

PRODUCER_TOKEN_COUNT = Gauge(
    "gabriel_producer_token_count",
    "Number of tokens remaining at each producer",
    ["producer_id"],
)

CLIENT_INPUTS_SENT_TOTAL = Counter(
    "gabriel_producer_inputs_sent_total",
    "Total number of client inputs sent from a producer",
    ["producer_id"],
)

CLIENT_INPUT_PROCESSING_LATENCY = Histogram(
    "gabriel_client_input_processing_latency_seconds",
    "End-to-end client input processing latency",
    ["producer_id"],
)


class InputProducer:
    """An input producer that produces inputs to send to the server.

    Stores a callback for producing input frames as well as the identifiers
    of the cognitive engines to target.

    The input producer can be stopped and resumed using :meth:`stop` and
    :meth:`resume`. The current status (stopped/running) can be obtained using
    :meth:`is_running`. By default, the input producer is running.

    The methods of this class are thread-safe.
    """

    def __init__(
        self,
        producer: Callable[[], Coroutine[Any, Any, InputFrame | None]],
        target_engine_ids: Iterable[str],
        producer_name: Union[str, None] = None,
    ):
        """Initialize the input producer.

        Args:
            producer (coroutine function):
                A coroutine function that produces input data
            target_engine_ids (Iterable[str]):
                Target engine IDs for the input
            producer_name (str, optional):
                The name of the producer producing the input
        """
        self._running = threading.Event()
        self._running.set()
        self._producer = producer
        self._target_engine_ids = set(target_engine_ids)
        self._target_engine_lock = threading.Lock()
        self.producer_name = producer_name
        self.producer_id = (
            producer_name + "-" + str(uuid.uuid4())
            if producer_name
            else str(uuid.uuid4())
        )
        self._loop = None

    async def produce(self) -> InputFrame | None:
        """Invoke the producer to generate input.

        Raises:
            Exception: if the producer is not running
        """
        if not self._running.is_set():
            raise Exception(
                f"Producer {self.producer_name} called when not running"
            )
        res = await self._producer()
        return res

    def resume(self) -> None:
        """Resume the producer."""
        if self._running.is_set():
            raise Exception("Producer already started")
        self._running.set()
        with self._target_engine_lock:
            logger.info(
                f"Resuming producer {self.producer_name} and targeting "
                f"engines {self._target_engine_ids}"
            )

    def stop(self) -> None:
        """Stop the producer."""
        logger.info("Stopping producer")
        self._running.clear()

    def change_target_engines(self, target_engine_ids: Iterable[str]) -> None:
        """Change the target engines for the producer.

        Args:
            target_engine_ids (list[str]):
                A list of target engine IDs for the input

        """
        with self._target_engine_lock:
            self._target_engine_ids = set(target_engine_ids)
        logger.info(
            f"Changing target engines to {target_engine_ids} for "
            f"producer {self.producer_name}"
        )

    def add_target_engine(self, target_engine_id: str) -> None:
        """Add a target engine to producer.

        Args:
            target_engine_id (str): a target engine id to add

        """
        with self._target_engine_lock:
            self._target_engine_ids.add(target_engine_id)
        logger.info(
            f"Adding {target_engine_id} to target engine ids for producer "
            f"{self.producer_name}"
        )

    def remove_target_engine(self, target_engine_id: str) -> bool:
        """Remove a target engine from the producer.

        Args:
            target_engine_id (str): a target engine id to remove

        Returns:
            True if the target engine id exists and was removed
            False if the target engine id does not exist

        """
        with self._target_engine_lock:
            if target_engine_id not in self._target_engine_ids:
                return False
            self._target_engine_ids.remove(target_engine_id)
            return True

    def is_running(self) -> bool:
        """Check if the producer is running."""
        return self._running.is_set()

    def _wait_for_running_internal(self, shutdown_event):
        while not shutdown_event.is_set():
            if self._running.wait(0.1):
                break

    async def wait_for_running(self) -> None:
        """Block until the producer is running."""
        shutdown_event = threading.Event()
        try:
            await asyncio.get_running_loop().run_in_executor(
                None, self._wait_for_running_internal, shutdown_event
            )
        except asyncio.CancelledError:
            shutdown_event.set()
            raise

    def get_target_engines(self) -> frozenset[str]:
        """Return the target engines for the producer."""
        with self._target_engine_lock:
            return frozenset(self._target_engine_ids)


class TokenPool:
    """A pool of tokens.

    Used to limit the number of in-flight requests for a particular
    input source.
    """

    def __init__(self, num_tokens: int, producer_id: str):
        """Initialize the token pool.

        Args:
            num_tokens (int): The number of tokens in the pool
            producer_id (str): The producer identifier

        """
        self._max_tokens = num_tokens
        self._sem = asyncio.BoundedSemaphore(num_tokens)
        self._producer_id = producer_id
        PRODUCER_TOKEN_COUNT.labels(producer_id=producer_id).set(
            self.get_remaining_tokens()
        )

    def return_token(self) -> None:
        """Return a token to the pool."""
        PRODUCER_TOKEN_COUNT.labels(producer_id=self._producer_id).inc()
        self._sem.release()

    async def get_token(self) -> None:
        """Acquire a token from the pool.

        Wait if necessary until a token is available.
        """
        logger.debug("Waiting for token")
        await self._sem.acquire()
        PRODUCER_TOKEN_COUNT.labels(producer_id=self._producer_id).set(
            self.get_remaining_tokens()
        )
        logger.debug("Token acquired")

    def is_locked(self) -> bool:
        """Check if the semaphore is locked."""
        return self._sem.locked()

    def reset_tokens(self) -> None:
        """Reset the number of tokens in the pool to the max number."""
        self._sem = asyncio.Semaphore(self._max_tokens)
        PRODUCER_TOKEN_COUNT.labels(producer_id=self._producer_id).set(
            self.get_remaining_tokens()
        )

    def get_remaining_tokens(self) -> int:
        """Return the number of remaining tokens in the pool."""
        return self._sem._value


class GabrielClient(ABC):
    """Abstract base class for a Gabriel client."""

    def __init__(self, prometheus_port: int):
        """Initialize the Gabriel client.

        Args:
            prometheus_port (int): Port for Prometheus metrics.
        """
        self._running = True
        # Whether a welcome message has been received from the server
        self._welcome_event = asyncio.Event()
        self.input_producers = set()
        # The number of tokens per input source, as specified by the \
        # server
        self._num_tokens_per_producer = None
        # Mapping from source id to tokens
        self._tokens = {}
        # Mapping from frame ids to the timestamp at which the input was
        # sent to the server
        self._pending_results = {}
        self._prometheus_port = prometheus_port

    def launch(self) -> None:
        """Launch the client synchronously.

        This method will block execution until the client is stopped.
        """
        asyncio.run(self.launch_async())

    @abstractmethod
    def launch_async(self) -> None:
        """Launch the client asynchronously."""
        pass

    def stop(self) -> None:
        """Stop the client."""
        self._running = False

    def record_send_metrics(self, from_client: FromClient) -> bool:
        """Record metrics related to sending of input to server."""
        producer_id = from_client.producer_id
        CLIENT_INPUTS_SENT_TOTAL.labels(producer_id=producer_id).inc()

        frame_id = from_client.frame_id
        self._pending_results[frame_id] = time.monotonic()

    def record_response_latency(self, result_wrapper: ToClient.ResultWrapper):
        """Record the response latency for input."""
        producer_id = result_wrapper.producer_id
        result = result_wrapper.result
        frame_id = result.frame_id

        send_time = self._pending_results[frame_id]
        latency = time.monotonic() - send_time

        CLIENT_INPUT_PROCESSING_LATENCY.labels(
            producer_id=producer_id
        ).observe(latency)



"""Push-based source that can be used to send frames to Gabriel server."""

import asyncio
import multiprocessing

from gabriel_protocol import gabriel_pb2

from gabriel_client.gabriel_client import InputProducer


def consumer(_):
    """A consumer that does nothing."""
    pass


class Source:
    """A push-based source used to send frames to Gabriel server."""

    def __init__(self, source_name, target_engine_ids):
        """Initialize the push-based source."""
        self._source_name = source_name
        self._frame_available = asyncio.Semaphore(0)
        self._latest_input_frame = None
        self._read, self._write = multiprocessing.Pipe(duplex=False)
        self._added_callback = False
        self._target_engine_ids = target_engine_ids

    def get_input_producer(self):
        """Returns an input producer for the source."""

        def reader_callback():
            input_frame = gabriel_pb2.InputFrame()
            input_frame.ParseFromString(self._read.recv_bytes())
            self._latest_input_frame = input_frame
            self._frame_available.release()

        async def producer():
            if not self._added_callback:
                # We need this to be run on the event loop running the producer
                fd = self._read.fileno()
                asyncio.get_event_loop().add_reader(fd, reader_callback)
                self._added_callback = True

            await self._frame_available.acquire()
            return self._latest_input_frame

        return InputProducer(
            producer=producer,
            target_engine_ids=self._target_engine_ids,
            source_name=self._source_name,
        )

    def send(self, input_frame):
        """Sends an input frame to the Gabriel server."""
        self._write.send_bytes(input_frame.SerializeToString())
